library IEEE;
use IEEE.STD_LOGIC_1164.all;
--use IEEE.STD_LOGIC_ARITH.all;
--use IEEE.STD_LOGIC_SIGNED.all;
use IEEE.numeric_std.all;

-------------------------------------------------------------------------------
-- entity
-------------------------------------------------------------------------------

entity de2_audio_out is
port (
    clk : in std_logic;                                -- By this here we mean the 100Mhz clock (updated from 50Mhz)
    reset_n : in std_logic;

    -- Bus master signals
    address      : in std_logic_vector (7 downto 0);
    byteenable   : in std_logic_vector (1 downto 0);
    writedata    : in std_logic_vector (15 downto 0);
    read         : in std_logic;
    write        : in std_logic;
    chipselect   : in std_logic;

     -- Slave signals
    readdata     : out std_logic_vector (15 downto 0);
    waitrequest  : out std_logic;   
   
    -- Audio interface signals
    AUD_CLK         : in std_logic;                  --    18.43MHz audio clock AUD_XCK
    AUD_DACLRCK  : out std_logic;                    --    Audio CODEC DAC LR Clock
    AUD_DACDAT   : out std_logic;                    --    Audio CODEC DAC Data
    AUD_BCLK     : inout std_logic;                  --    Audio CODEC Bit-Stream Clock
    AUD_XCK      : out std_logic;
    -- Test signals --
    ledr    : out std_logic_vector (17 downto 0)
  );
end de2_audio_out;

-------------------------------------------------------------------------------
-- architecture
-------------------------------------------------------------------------------

architecture imp of de2_audio_out is

      component de2_fifo
      port (
            data        : IN STD_LOGIC_VECTOR (15 DOWNTO 0);
            rdclk        : IN STD_LOGIC ;
            rdreq        : IN STD_LOGIC ;
            wrclk        : IN STD_LOGIC ;
            wrreq        : IN STD_LOGIC ;
            q            : OUT STD_LOGIC_VECTOR (15 DOWNTO 0);
            rdempty        : OUT STD_LOGIC ;
            wrfull        : OUT STD_LOGIC ;
            wrusedw        : OUT STD_LOGIC_VECTOR (8 DOWNTO 0)
      );
      end component;
  
    component de2_wm8731_audio_out is
    port (
        clk : in std_logic;       --  Audio CODEC Chip Clock AUD_XCK (18.43 MHz)
        reset_n : in std_logic;
        test_mode : in std_logic;       --    Audio CODEC controller test mode
        data_in : in std_logic_vector(15 downto 0);   
        audio_req : out std_logic;
     
        -- Audio interface signals
        AUD_DACLRCK  : out std_logic;   --    Audio CODEC DAC LR Clock
        AUD_DACDAT   : out std_logic   --    Audio CODEC DAC Data
    );
    end component;   

    signal reset : std_logic;

    signal dac_request : std_logic;
    signal adc_request : std_logic;
    signal data_from_bus : std_logic_vector (15 downto 0);
    signal data_to_bus : std_logic_vector (15 downto 0);
    signal adc_data : std_logic_vector (15 downto 0);
    signal dac_data : std_logic_vector (15 downto 0);
	signal adc_rdempty : std_logic;
    signal adc_wrfull : std_logic;
    signal dac_rdempty : std_logic;
    signal dac_wrfull : std_logic;
    signal writefifo   : std_logic;
    signal readfifo       : std_logic;   
    signal adc_rdreq   : std_logic;   
    signal dac_wrreq   : std_logic;   
   
   
    --signal counter : std_logic_vector (13 downto 0);
    --signal sin_counter : std_logic_vector (5 downto 0);
    --signal sin_out : std_logic_vector (15 downto 0);

    --signal rate : std_logic_vector(2 downto 0); -- "XX1" 11kHz, "01X" 22kHz, "1XX" 44kHz
    --signal rate_counter : std_logic_vector (2 downto 0);
    --signal req_fifo : std_logic;

    signal adc_used_buf : std_logic_vector (8 downto 0);
    signal dac_used_buf : std_logic_vector (8 downto 0);
    signal adc_before_buf : std_logic_vector (8 downto 0);
    signal adc_after_buf : std_logic_vector (8 downto 0);
    signal dac_before_buf : std_logic_vector (8 downto 0);
    signal dac_after_buf : std_logic_vector (8 downto 0);
    signal adc_stop : std_logic;
    signal dac_stop : std_logic;

    --test
    signal ledr_count : std_logic_vector (16 downto 0);
    signal led_ctrl : std_logic;
   
    --Audio clk dividers
    signal sampling_clk : std_logic;
    signal audio_counter : unsigned (13 downto 0);
    signal counter_aud_adclkrck : unsigned (15 downto 0);
    signal LRCK: std_logic;
    signal bus_state: std_logic_vector (2 downto 0);

	type state_t is (a0, a1, a2, d0, d1, d2);
	signal adc_state, adc_new_state, dac_state, dac_new_state : state_t;
   
begin

--Turn off the Auto RAM Replacement or Auto Shift Register Replacement logic options
--for the project in the Analysis & Synthesis Settings page of the Settings dialog box
--(Assignments menu), or to specific entities using the Assignment Editor (Assignments menu). 
--Current Global Settings: Auto RAM Replacement = ON  (Recommended: OFF)
--Auto Shift Register Replacement = AUTO  (Recommended: OFF)

    dac_wrreq <= writefifo;
   
	SM_DAC: process(dac_state)
	begin
		case dac_state is
			when d0 =>
				if (dac_wrfull = '0' and write = '1') then
					dac_new_state <= d1;
				else
					dac_new_state <= d0;
				end if;					
			when d1 =>
				if (dac_stop = '1') then
					dac_new_state <= d2;
				else
					dac_new_state <= d1;
				end if;
				
			when d2 =>
				if (write = '1' and dac_wrfull = '0') then
					dac_new_state <= d1;
				else 
					dac_new_state <= d0;
				end if;					
			when others =>
				dac_new_state <= d0;
		end case;
	end process SM_DAC;	
	
	ST_DAC: process(clk, reset_n)
	begin
		if reset_n = '0' then
			dac_state <= d0;
		elsif rising_edge(clk) then
			if (chipselect = '1') then
				dac_state <= dac_new_state;
			end if;
		end if;
	end process ST_DAC;
	
    AudioCounter: process (clk)
    begin
      if rising_edge (clk) then
        if reset = '1' then
            audio_counter <= (others => '0');
        elsif audio_counter = X"30D3" then -- 12500 -1 = 12499
            audio_counter <= (others => '0');
        else
            audio_counter <= audio_counter + 1;
        end if;
      end if;
    end process AudioCounter;
    
    SamplingClock: process (clk)
    begin
      if rising_edge (clk) then
        if reset = '1' then
            sampling_clk <= '0';
        elsif audio_counter = X"30D3" then -- 12500 -1 = 12499
            sampling_clk <= not sampling_clk;
        end if;
      end if;
    end process SamplingClock;

	BusComm: process(clk, reset_n)
	begin
		if reset_n = '0' then
			readdata <= (others => '0');
            waitrequest <= '0';	
			adc_stop <= '0';
			dac_stop <= '0';
		elsif rising_edge(clk) then
			if dac_new_state = d0 then
				if chipselect = '1' then
					writefifo <= '0';
					waitrequest <= '0';
					dac_stop <= '0';
				end if;
			elsif dac_new_state = d1 then
				if chipselect = '1' then
					if (address(0) = '1' and write = '1') then
						data_from_bus <= writedata;   
						writefifo <= '1';
						waitrequest <= '1';   
						dac_stop <= '1';
					end if;
				end if;
			elsif dac_new_state = d2 then
				if chipselect = '1' then
					waitrequest <= '0';
					dac_stop <= '0';	
				end if;			
			end if;
			
		end if;
	end process BusComm;
     
    process(AUD_BCLK)
    begin
    if rising_edge(AUD_BCLK) then
        counter_aud_adclkrck <= counter_aud_adclkrck + 1;
    end if;

    if (counter_aud_adclkrck = "0000000000000000") then
        led_ctrl <= not led_ctrl;
    end if;
    end process;
   
    AUD_XCK <= AUD_CLK;

    DAC : de2_wm8731_audio_out
    port map (
        clk => AUD_CLK,
        reset_n => reset_n,     
        test_mode => '0',
        audio_req => dac_request,
        data_in => dac_data,         

        AUD_DACLRCK  => AUD_DACLRCK,
        AUD_DACDAT   => AUD_DACDAT
    ); 

    DAC_FIFO : de2_fifo
    port map (
        data        => data_from_bus,
        rdclk        => AUD_CLK,
        rdreq        => dac_request,
        wrclk        => clk,
        wrreq        => dac_wrreq,
        q            => dac_data,
        rdempty        => dac_rdempty,
        wrfull        => dac_wrfull,
        wrusedw        => dac_used_buf   
    );   
   
--      ledr(5)  <= led_ctrl;
--      ledr(6)  <= dac_wrfull;
--      ledr(7)  <= adc_rdempty;
--      ledr(4)  <= '0';

end architecture imp;